# pio_buzzer_3.py - 12.09.2021
#
# siehe auch pio_buzzer_1.py
# Die Kommunikation zwischen dem Taktgeber und dem Tongenerator
# erfolgt über (genaugenommen 2) Pins:
# Der Taktgeber signalisiert mit einem High, dass der Tongenerator
# aktiv werden soll.
# Dieser Pin kann auch zur optischen Anzeige des Signals mittels
# LED genutzt werden.
# Die Ein-Auszeiten werden durch ein 32-Bit-Pattern bestimmt,
# das über einen Zeitraum von ca 1 Sekunde abläuft.
# default = 0x0000_FFFF -> 1/2 Sek. Beep, 1/2 Sek. Pause
#           0x00FF_00FF -> 1/4 Sek. Beep, 1/4 Sel. Pause  ...
# Das höchstwertige Bit soll 0 sein, damit am Summer am Ende
# keine Spannung mehr anliegt.
    
import time
from rp2 import PIO, StateMachine, asm_pio
from machine import Pin

class Buzzer(object):
    def __init__(self, buzzer_pin, synch_out_pin, synch_in_pin, buzzer_freq=3250):
        self.sm0 = StateMachine(0, self.buzzer, freq=buzzer_freq*2,
                                sideset_base=Pin(buzzer_pin),
                                in_base=Pin(synch_in_pin, Pin.IN, Pin.PULL_DOWN) )
                                
        self.sm1 = StateMachine(1, self.cycle,
                                out_base=Pin(synch_out_pin),
                                freq=2000)
        self.sm0.active(1)
        self.sm1.active(1)
        
    def beep(self, cnt=1, pattern=0x0000_FFFF):
        # ggf. eine Pause einlegen
        if cnt < 1:
            self.sm1.put(0)
        else:
            # das höchstwerige Bit löschen
            pattern &= 0x7FFF_FFFF
            for _ in range (cnt):
                self.sm1.put(pattern)
    
    def tx_free(self):
        # Anzahl der freien Words im RX-Fifo
        return 8 - self.sm1.tx_fifo()
    
    @staticmethod
    @asm_pio(sideset_init=PIO.OUT_LOW)
    # Hier wird der Ton generiert = 1/2 des F_PIO-Taktes
    def buzzer():
        wrap_target()
        
        nop()            .side(1)  #  1
        wait(1, pin, 0)  .side(0)  #  2 Warten bis Pin auf High
        wrap()
        
    @staticmethod
    @asm_pio(out_shiftdir=PIO.SHIFT_RIGHT,
             out_init=PIO.OUT_LOW,
             fifo_join=PIO.JOIN_TX)
    # hier werden Pulse und Pause generiert,
    # die den Buzzer ein- und ausschalten.
    def cycle():
        wrap_target()
        pull(block)                 #  1 blocking warten auf FIFO
        
        label("next")               #  2 
        out(pins, 1)  [31]          #  3 jeweils 1 Bit nach x ausgeben
        jmp(not_osre, "next") [31]  #  4 alle 32 Bit rausshiften
        
        wrap()                      #    dann das nächste Word holen


if __name__ == "__main__":
    buzzer_pin    = 19        # Summer
    synch_out_pin = 20        # Ausgang Takt
    synch_in_pin  = 21        # Eingang Tongenerator
    
    buzz = Buzzer(buzzer_pin=buzzer_pin,
                  synch_out_pin=synch_out_pin,
                  synch_in_pin=synch_in_pin)

    buzz.beep()
    buzz.beep(0)
    buzz.beep(1, 0x3333_3333)
    buzz.beep(1, 0x1248_00FF)
    buzz.beep(1, 0x1248_1248)

    
    print("\ndone")
