# pio_ir_nec_rx.py  -  12.09.2021
#
# IR-Remote-Receiver für das NEC-Protokoll

import uarray as arr
from rp2 import PIO, StateMachine, asm_pio
from machine import Pin, lightsleep

# ---------------------------------------

@asm_pio(in_shiftdir=PIO.SHIFT_RIGHT,
             fifo_join=PIO.JOIN_RX)

# Takt    : 50us / 20_000 Hz
# Startbit  9000 + 4500 us 
# Bit_0    = 560 +  560 us  
# Bit_1    = 560 + 1690 us 
# Stop     = 560 + 30ms Wartezeit
# [Wiederholung: 8000us + 2250 us + 560 us + 100 ms Pause]

def ir_rx():
    wrap_target()

    label("restart")            #    Ruheleel high, warten auf Startbit-Pulse -> low
    wait(0, pin, 0)             #  1 warten, bis der Pin auf low 
                                #    8500us warten = 170 * 50
                                #    In dieser Zeit darf der Pegel nicht auf high wechsel
    set(x, 28)                  #  2 wenn doch, dann liegt kein Startbit vor
    label("loop_1")

    jmp(pin, "restart")         #  3 wenn der Pin in dieser Zeit auf high geht, 
                                #    dann liegt kein korrektes Startbit vorm
    jmp(x_dec, "loop_1")   [4]  #  4 (28 + 1) * (1 + 4 + 1) * 50 = 29 * 6 * 50 = 174 * 50  = 8700
   
                                #    Jetzt sind 8700us um, der Pegel des Startbots ist immer noch low                         
    set(x, 9)                   #  5 Warteschleife 1000us, in diesem Intervall muss der Pin auf high gehen
    label("loop_2")
    jmp(pin, "pause_start")     #  6 wenn der Pin auf high geht, dann weiter
    jmp(x_dec, "loop_2")        #  7 (9 + 1) * (1 + 1) * 50 = 20 * 50 = 1000

    jmp("restart")              #  8 Die Wartezeit ist um, der Pegel ist immer low,
                                #    das Startbit ist nicht korrekt
                                
    label("pause_start")        #    Die Pause hat rechtzeitig begonnen

    set(x, 19)                  #  9 warten 4000us
                                #    In dieser Zeit darf der Pegel nicht auf low gehen
    label("loop_3")
    jmp(pin, "loop_3a")         # 10 wenn der Pin noch high ist, dann weiter in der Schleife
    jmp("restart")              # 11 wenn dier Pegel low ist, dann Fehler in Pause des Startbits
    
    label("loop_3a")
    jmp(x_dec, "loop_3")   [2]  # 12 (19 + 1) * (1 + 2 + 1 ) * 50 = 80 * 50 = 4000

                                #    4000us Pause sind um, der Pegel ist immer noch high
                                #    nun 1000us warten auf den Pegelwechsel nach low
    set(x, 9)                   # 13 In diesem Intervall muss der Pegel auf low gehen
    label("loop_4")
    jmp(pin, "loop_4a")         # 14 der Pin istimmer noch high, weiter in der Warteschleife
    jmp("read_data")            # 15 der Pin ist low, weiter zum Einlesen der Datenbits
    label("loop_4a")
    jmp(x_dec, "loop_4")        # 16 (9 + 1) * (1 + 1) * 50 = 20 * 50 = 1000

    jmp("restart")              # 17 Die Wartezeit ist abgelaufen, der Pegel ist immer noch high
                                #    Die Pause des Startbits ist nicht korrekt
                                    
    label("read_data")          #    Das Startbit war ok, jetzt Datenbits lesen, der Pegel ist low    
    set(y, 31)                  # 18 Zähler für die zu empfangenden Bits setzen

    label("next_bit")           #    Der Pegel ist low, mindestens 1 steigende Flanke muss und wird kommen,
                                #    sobald das IR-Signal endet
    wait(1, pin, 0)        [17] # 19 Zunächst Warten, bis der Pin von low wieder auf high geht
                                #    danach noch eimal warten 900us = (17 + 1) * 50 = 18 * 50 = 900

    in_(pins, 1)                # 20 Den Pegel des InputPins ins ISR kopieren
                                #    Wenn low  (kurze Pause), dann wird eine 0 übertragen, 
                                #    wenn high (lange Pause), dann eine 1

                                #    Der Pegel ist jetzt low (wenn Datenbit 0) oder high bei Datenbit 1)
                                
    jmp(pin, "loop_5a")         # 21 Wenn high, dann auf Flankenwechsel nach low warten
    jmp("end_bit")              # 22 wenn low, dann auf die nächste fallende Flanke warten
    
    label("loop_5a")
    set(x, 10)                  # 23
    label("loop_6")             # 24  Wartezeit > 560us = (5 + 1) * (1 + 1) * 50us = 500us
    jmp(pin, "loop_6a")         # 25 Der Pin ist noch high, weiter in der Zeitschleife
    jmp("end_bit")              # 26 Der PIn ist low, das nächste Bit einlesen
    
    label("loop_6a")
    jmp(x_dec, "loop_6")        # 27 600us auf low warten (10 + 1 + 1) * 50 = 12 * 50
    
    jmp("restart")              # 28 Wartezeit abgelaufen, der Pin ist immer noch high
                                #    entweder war es das Stopbit oder ein Fehler
    label("end_bit")
    jmp(y_dec, "next_bit")      # 29 Bitcounter
                                #    32 Bit sind jetzt eingelesen

    push()                      # 30 das Ergebnis ins Fifo und isr zurücksetzen
    irq(rel(0))                 # 31 Interrupt auslösen, damit die Daten aus dem Fifo abgeholt werden

    wrap()
   
# --------------------------------------------------------

class IR_RX(object):
    def __init__(self, ir_pin):
        self.sm = StateMachine(0, ir_rx,
                in_base=Pin(ir_pin, Pin.IN, Pin.PULL_UP),
                jmp_pin=Pin(ir_pin, Pin.IN, Pin.PULL_UP),
                freq=20_000)
        self.data = arr.array("B", [0] * 4)
        self.recv = 0
        self.sm.irq(self.irq_handler)
        self.sm.restart()
        self.sm.active(1)
   
    def irq_handler(self, _):
        tmp = self.sm.get()
        check_sum = 0
        for i in range(4):
            self.data[i] = tmp & 0xFF
            check_sum += self.data[i]
            tmp >>= 8
        self.recv = 1 if check_sum == 255 * 2 else -1
    
    def received(self): 
        return self.recv
    
    def get_data(self):
        self.recv = 0
        return self.data[1], self.data[2]  # adr, cmd
    
if __name__ == "__main__":

    rx_pin = 17
    tx = IR_RX(ir_pin=rx_pin)
    
    print("waiting for IR-signal ... \n") 

    while True:
       lightsleep(50)
       
       if tx.received():
           if tx.received() < 0:
               print("Error checksum ! ", end = "")
           adr, cmd = tx.get_data()
           print("Adr: {}, CMD: {}".format(adr, cmd))
 