# pio_uart_rx_4b.py - 12.09.2021
#
# Wie pio_uart_rx_4a.py, jedoch wird hier das Ende der Übertragung festgestellt
# und es kann eine beliebige Anzahl von Bytes empfangen werden.
#
# Prüft das Startbit 2x, an der Stelle des Stopbits muss der Pegel high sein.
# Das Ende der Übertragung wird erkannt, wenn nach dem letzten Byte innerhalb
# von 1 Takt kein neues Startbit empfangen wird.
#
# Jeweils nach dem Empfang von 4 Bytes wird ein Interrupt ausgelöst,
# der nterrupt-Handler liest das Fifo-Word in einen Ring-Buffer von 16 Words ein.
# Wurde das Ende der Übertragung erkannt, dann werden die bis dahin empfangenen
# Bytes (maximal 3 - sonst wären die vorher bereits ins Fifo gepushed worden)
# in ein Fifo geschrieben.
# Zus#tzlich eird in einem 2.Word als Endemarkierung die Anzahl der Bytes im vorletzten
# Fifo-Buffer geschrieben.
#
# Beim Auslesen der Fifos wird unterschieden zwischen dem
# - letzten Fifo (dort steht, wieviele empfangene Bytes das vorletzte Fifo enthält)
# - vorletzten Fifo (dort sind zwischen 0 und 3 Bytes an Daten enthalten) und den
# - vorausgegangenen Fifos (dort sind immer 4 Bytes an empfangenen Daten enthalten).

from machine import Pin, lightsleep
import uarray as arr
from rp2 import PIO, StateMachine, asm_pio
import time

# ----------------------------------------------------------------------------
# Diese Version prüft das Startbit 2x, das Stopbit und die nachfolgende Pause
# Jedes Bit hat eine Länge von 15 PIO-Takten.
# Die Datenbits werden jeweils 1x in der Bitmitte abgetastet.
# Takt = Baud * 15

@rp2.asm_pio(
        #sideset_init=rp2.PIO.OUT_HIGH, 
        in_shiftdir=rp2.PIO.SHIFT_RIGHT,
        fifo_join=rp2.PIO.JOIN_RX ) 

def uart_rx():
    label("restart")  
    mov(isr, null)            #  1 Bei einem Fehler sicherheitshalber das ISR zurücksetzen
    
    wait(0, pin, 0)           #  2 Warten auf eine fallende Flanke
    set(x, 3)                 #  3 Zähler für die 4 Bytes pro Fifo

    label("start")    
    wrap_target()
    set(y, 7)                 #  4 den Counter fpr die Bits setzen
                              #    Das Signal ist mindestens 4 Takte gesetzt
    jmp(pin, "restart")   [6] #  5 - Fehler, wenn der Takt das Startbit nicht mehr low 
    jmp(pin, "restart")   [5] #  6 - Fehler, wenn der Takt das Startbit nicht mehr low 
    nop()                 [6] #  7 warten bis die Bit-Mitte des 1. Bits erreicht ist

    label("next_bit")
    in_(pins, 1)          [6] #  8 Den Pegel einlesen und nach
    nop()   [6]               #  9
    jmp(y_dec, "next_bit")    # 10 weiteren 15 Takten das nächste Bit einlesen
    
    jmp(pin, "stop_bit_1")    # 11 Das Stopbit muss high, sein 
    jmp("restart")            # 12 wenn nicht gesetzt, dann ist das ein Fehler
    
    label("stop_bit_1")
    jmp(pin, "byte_end")      # 13 Das Stopbit muss high sein, 
    jmp("restart")            # 14 wenn nicht gesetzt, dann ist das ein Fehler                         
    
    label("byte_end")
                   
    jmp(x_dec, "end_1")       # 15 erst nach 4 Bytes pushen

    push()                    # 16 die empfangenen Bits ins Fifo
    set(x, 3)                 # 17 den Zähler für die Bytes neu setzen
    irq(rel(0))               # 18 Interrupt auslösen nach jeweils 4 Byte   
    
    label("end_1")
    set(y, 4)                 # 19 Wartezeit nach dem Stopbit nach 'y'
                              #    wenn innerhalb dieser Zeit der Level nicht
                              #    azf gnd geht, dann ist die Übertragung beendet
    label("end_2")
    jmp(y_dec, "end_3")       # 20 Zähler Wartezeit decrementieren
    jmp("read_end")           # 21
    
    label("end_3")            #
    jmp(pin, "end_2")         # 22 hier prüfen, ob eine fallende Flanke auftrat 
    wrap()                    #    wenn ja, dann das nächste Byte einlesen
 
    label("read_end")
    push()                    # 23 den letzten Datensatz pushen
    mov(isr, x)               # 24 das X-Register pushen
    push()                    # 25 Der irq-handler sieht nun 2 neue Words im Fif0
    irq(rel(0))               # 26 den Interrupt auslösen
    
    jmp("restart")            # 26
    
class UART_RX(object):
    def __init__(self, statemachine, rx_pin, baud=9600, read_4_bytes=False):
        self.sm = StateMachine(statemachine, uart_rx, 
                in_base=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
                jmp_pin=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
                #sideset_base=Pin(18),
                freq=baud * 15)
        
        self.sm.restart()
        self.sm.irq(self.irq_handler)
        self.rx_word_buffer = arr.array("I", [0] * 16)
        # Buffer um jeweils 4 Byte (1 Fifo-Word) abzuholen
        self.byte_buffer = arr.array("B", [0] * 4)
        self.rx_tail = 0
        self.rx_head = 0
        self.last_word = 0
        self.last_cnt = 0
        self.rx_complete = False

    def activate(self, state=1):
        self.sm.active(1 if state else 0)
        
    def irq_handler(self, _):
        # irq-handler zum Einlesen der empfangenen Bytes
        # Am Ende der Übertragung erfolgen 2 push() von der Statemachine
        # Am Eingang von 2 Words kann daher der IRQ-Handler das Ende erkennen.

        # Am Ende wird ein Datensatz mit der Anzahl der Bytes
        # im vorletzten Datensatz gesendet.
        if self.sm.rx_fifo() == 2:
            self.rx_complete = True
            self.last_word = self.sm.get()
            self.last_cnt = self.sm.get()
        else:
            self.rx_word_buffer[self.rx_head] = self.sm.get()
            self.rx_head = (self.rx_head + 1) & 0x0F
            
            #if self.rx_head == self.rx_tail:
            #    raise IndexError("Head overwrites Tail")
            
    def complete(self):
        return self.rx_complete
    
    def received(self):
        # Wenn rx_head == rx_tail, dann gibts nicht auszulesen
        return not (self.rx_head == self.rx_tail)
    
    def get_data(self):
        # Genau ein Fife-Word in Bytes umkopieren und zurückgeben
        for i in range(4):
            self.byte_buffer[i] = self.rx_word_buffer[self.rx_tail] & 0xFF
            self.rx_word_buffer[self.rx_tail] >>= 8 
        self.rx_tail = (self.rx_tail + 1) & 0x0F
        #print(self.rx_head,  self.rx_tail)
        return self.byte_buffer
        
    def get_last(self):
        # Das Ende der Übertragung ist erkann
        # In last_word steht der letzte Datensatz
        # in last_cnt steht, wieviele Bytes dort enthalten sind
        # last_cnt = X-Register am Ende der Übertragung
        
        # x = 3 - __ __ __ __
        # x = 2 - 11 __ __ __
        # x = 1 - 22 11 __ __
        # x = 0 - 33 22 11 __
        
        empty = self.last_cnt + 1
        while empty:           
            empty -= 1
            self.last_word >>= 8
            
        # die Datenbytes im letzten Word freistellen
        for i in range(4 - self.last_cnt):
            self.byte_buffer[i] = self.last_word & 0xFF
            self.last_word >>= 8
        return (3 - self.last_cnt), self.byte_buffer
 
    def clear(self):
        self.rx_complete = False
        self.rx_head = self.rx_tail = 0
          
if __name__ == "__main__":
    
    # -- Einstellung ------------------------
    
    BAUD = 9600 
    rx_pin = 17        
      
    # ----------------------------------------
    
    rx = UART_RX(statemachine=0, rx_pin=rx_pin, baud=BAUD)
    rx.activate()
    
    # Byte-Buffer um die gesamte Sendung aufzunehmen
    rx_byte_buffer = arr.array("B", [0] * 228)
    rx_idx = 0
    rx_cnt = 0
    
    print("Waiting for UART-Data ... Ctrl-C to stop.\n")
    try:
        while True:
            lightsleep(5)
            
            # ggf. mehrere Wordd abholen
            while rx.received():
                # liefert ein Array mit genau 4 Bytes zurück
                buffer = rx.get_data()
                for i in range(4):
                    rx_byte_buffer[rx_idx] = buffer[i]
                    rx_idx += 1
                     
            # Am Ende der Übertragung das letzte Word abholen        
            if rx.complete():
                # liefert ein Array zurock mit cnt Bytes [0 - 3]
                cnt, buffer  = rx.get_last()
                
                for i in range(cnt):
                    rx_byte_buffer[rx_idx] = buffer[i]
                    rx_idx += 1
                
                print(bytes(rx_byte_buffer[:rx_idx]))
                print(rx_idx)
                
                rx.clear()
                rx_idx = 0

            
    except KeyboardInterrupt:
        pass
        #print(rx_byte_buffer)
        #print(rx_data)
        #print("Bytes received: {}, Datensätze: {}".format( rx_cnt, ready))
        
    rx.activate(0)
    print("\ndone")
